import os, sys, math, random, time, pygame
from pygame.locals import *
from pygame.compat import geterror

import const, util, image, var, item, bonus

class Ball(pygame.sprite.DirtySprite):
    def __init__(self, on_paddle, pos=[0,0], vel=[0,0]):
        pygame.sprite.DirtySprite.__init__(self)
        self.image_normal = image.images.ball_normal
        self.image_fire = image.images.ball_fire
        self.image = self.image_normal
        self.rect = self.image.get_rect()
        self.radius = self.rect[2] / 2
        self.vel_init = [2, -3]
        self.wall_hits = 0  #for a workaround: count of the consecutive hitting of walls and bricks (not paddle) to avoid the ball stuck in a loop
        self.on_paddle = on_paddle
        self.paddle_offset = 0
        if on_paddle:
            self.put_on_paddle()
        else:
            self.vel = vel
            self.pos = pos
            self.rect[0], self.rect[1] = self.pos[0] - self.radius, self.pos[1] - self.radius

    def fire(self, use):
        if use:
            self.image = self.image_fire
        else:
            self.image = self.image_normal

    def put_on_paddle(self):
        self.vel = [0, 0]
        self.pos = [var.game.paddle.pos + self.paddle_offset, const.PADDLE_Y - var.game.paddle.height - self.radius ]
        self.rect[0], self.rect[1] = self.pos[0] - self.radius, self.pos[1] - self.radius
        
    def update(self):
        self.dirty = 1
        if self.on_paddle:
            self.put_on_paddle()
            return

        posx = self.pos[0] + self.vel[0]
        posy = self.pos[1] + self.vel[1]
        #make the ball bounce from the walls 
        if posx <= self.radius or posx >= const.SCREEN_WIDTH - self.radius:
            self.vel[0] = -self.vel[0]
            posx = self.pos[0] + self.vel[0]
            self.wall_hits += 1
        if posy <= const.SCREEN_Y0 + self.radius:
            self.vel[1] = -self.vel[1]
            posy = self.pos[1] + self.vel[1]
            self.wall_hits += 1
        elif var.game.bonusdata["net"]["cnt"] > 0 and posy >= const.SCREEN_HEIGHT - self.radius - var.game.net.rect[3]:
            posy = const.SCREEN_HEIGHT - self.radius - var.game.net.rect[3]
            self.vel[1] = -self.vel[1]
        elif posy >= const.SCREEN_HEIGHT + self.radius:
            var.game.balls.remove(self)
            var.game.allsprites.remove(self)
            if len(var.game.balls) == 0:
                var.game.lose_life()
            return
        elif pygame.sprite.collide_mask(self, var.game.paddle):
            var.game.clear_combo()
            self.wall_hits = 0
            if var.game.bonusdata["sticky"]["cnt"] > 0:
                sticky = True
                for b in var.game.balls:
                    if b.on_paddle:   #one sticky ball already exists, it's enough
                        sticky = False
                        break
                if sticky:
                    self.on_paddle = True
                    self.paddle_offset = posx - var.game.paddle.pos
                    self.put_on_paddle()
                    return
            
            self.vel[1] = -self.vel[1]
            self.vel[0] = self.vel[0] + var.game.paddle.friction * math.copysign(min(20,math.sqrt(abs(var.game.paddle.vel))), var.game.paddle.vel)
            posx = self.pos[0] + self.vel[0]
            posy = const.PADDLE_Y - var.game.paddle.height - self.radius

        else:
            r0, r1 = self.rect[0], self.rect[1]
            self.rect[0] = posx - self.radius
            self.rect[1] = posy - self.radius            
            found = self.check_collisions()
            if found:
                self.rect[0], self.rect[1] = r0, r1
                return

        self.pos[0], self.pos[1] = posx, posy 
        self.rect[0] = self.pos[0] - self.radius
        self.rect[1] = self.pos[1] - self.radius
        self.fix_loop()

    def fix_loop(self):   #the workaround to avoid the ball getting stuck in a loop, adjusting the velocity to change the direction
        if self.wall_hits > 3:
            if abs(self.vel[0]) > abs(self.vel[1]):
                self.vel[1] += math.copysign(0.4, self.vel[1])
            else:
                self.vel[0] += math.copysign(0.4, self.vel[0])
            self.wall_hits = 0
           
    def release(self):
        self.on_paddle = False
        self.vel[0] = self.vel_init[0] + var.game.paddle.friction * math.copysign(min(20,math.sqrt(abs(var.game.paddle.vel))), var.game.paddle.vel)
        self.vel[1] = self.vel_init[1]
        self.update()

    def check_collisions(self):
        found = False
        for brick in list(var.game.bricks):
            if not found and brick.collides(self):
                found = True
                dead = False
                
                if var.game.bonusdata["fire"]["cnt"] == 0:
                    diffx = brick.rect[0] + brick.rect[2]/2 - self.pos[0]
                    diffy = brick.rect[1] + brick.rect[3]/2 - self.pos[1]
                    angle = math.atan2(diffy, diffx)
                    angle1 = angle % (math.pi * 2)
                    #bouncing off bricks, direction depends on the angle
                    if (angle1 > math.pi/4 and angle1 < 3*math.pi/4) or (angle1 > 5*math.pi/4 and angle1 < 7*math.pi/4):
                        #print "change dir up-down"
                        self.vel[1] = -self.vel[1]
                    elif (angle1 > 3*math.pi/4 and angle1 < 5*math.pi/4) or (angle1 > 7*math.pi/4 or angle1 < math.pi/4):
                        #print "change dir right-left"
                        self.vel[0] = -self.vel[0]
                    else: #corner
                        #print "corner"
                        self.vel[0] = -self.vel[0]
                        self.vel[1] = -self.vel[1]
                    
                    if brick.max_hits == 0:
                        self.wall_hits += 1
                        self.fix_loop()
                    else:
                        self.wall_hits = 0
                        dead = brick.hit()
                else:
                    brick.hit()   #need to call the hit() in case it does some action (e.g. dynamite)
                    dead = True   #kill everything on the way, even walls
                if dead:
                    var.game.add_bonus(brick.rect[0] + brick.rect[2]/2, brick.rect[1] + brick.rect[3])
                    brick.die()    
                    var.game.update_score(1)
                    var.game.update_combo()

            else:
                brick.collided = False
        return found


 
